# Generated by Django 4.0.4 on 2022-05-26 09:14
# pylint: disable=invalid-name

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="RefreshTokenRotationLog",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "expires_at",
                    models.DateTimeField(
                        editable=False,
                        help_text="Date and time when the token expires.",
                    ),
                ),
                (
                    "title",
                    models.UUIDField(
                        editable=False,
                        help_text="Title this refresh token holds.",
                    ),
                ),
            ],
        ),
    ]
